/***************************************************************************
                          kscope.cpp  -  description
                             -------------------
    begin                : Fri May  9 09:16:06 IDT 2003
    copyright            : (C) 2003 by Elad Lahav
    email                : elad_lahav@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <kfiledialog.h>
#include <kmenubar.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <klineedit.h>
#include <kinputdialog.h>
#include <kxmlguifactory.h>
#include <kstatusbar.h>
#include <kaction.h>
#include "kscope.h"
#include "kscopeconfig.h"
#include "projectmanager.h"
#include "editortabs.h"
#include "fileview.h"
#include "filelist.h"
#include "querywidget.h"
#include "editormanager.h"
#include "cscopefrontend.h"
#include "ctagslist.h"
#include "newprojectdlg.h"
#include "openprojectdlg.h"
#include "preferencesdlg.h"
#include "dirscanner.h"
#include "querypage.h"
#include "calltreedlg.h"
#include "kscopepixmaps.h"
#include "progressdlg.h"
#include "projectfilesdlg.h"
#include "cscopemsgdlg.h"
#include "positionhistory.h"
#include "positionhistorydlg.h"

/**
 * Class constructor.
 * @param	pParent	The parent widget
 * @param	szName	The widget's name
 */
KScope::KScope(QWidget* pParent, const char* szName) :
	KParts::DockMainWindow(pParent, szName),
	m_pProgressDlg(NULL)
{
	QString sPath;

	// Load configuration
	Config().load();

	// Connect menu and toolbar items with the object's slots
	m_pEditorMenu = new QPopupMenu(this);
	initActions();
	
	// Create the initial GUI (no active part)
	setXMLFile("kscopeui.rc");
	createShellGUI();
	m_bShellGUICreated = true;

	// Create all child widgets
	initMainWindow();

	// Create control objects
	m_pProjMgr = new ProjectManager();
	m_pEditMgr = new EditorManager(this);
	m_pCscope = new CscopeFrontend(m_pMsgDlg);

	Pixmaps().init();

	// Open a file for editing when selected in the project's file list or the
	// file tree
	connect(m_pFileView, SIGNAL(fileRequested(const QString&)), this,
		SLOT(slotAddEditor(const QString&)));

	// Delete an editor page object after it is removed
	connect(m_pEditTabs, SIGNAL(editorRemoved(EditorPage*)),
		this, SLOT(slotDeleteEditor(EditorPage*)));
	
	// Set an editor as the active part whenever its owner tab is selected
	connect(m_pEditTabs, SIGNAL(editorChanged(EditorPage*, EditorPage*)),
		this, SLOT(slotChangeEditor(EditorPage*, EditorPage*)));

	// Report whether Cscope is running or not
	connect(m_pCscope, SIGNAL(running(bool)), this,
		SLOT(slotCscopeRunning(bool)));

	// Report whether Cscope is ready or working
	connect(m_pCscope, SIGNAL(stateChanged(bool)), this,
		SLOT(slotCscopeState(bool)));

	// Display a progress dialog for lengthy Cscope operations (e.g.,
	// building the database or searching for symbols)
	connect(m_pCscope, SIGNAL(progress(const QString&, int)), this,
		SLOT(slotProgress(const QString&, int)));

	// Display a file at a specific line when selected in a query list
	connect(m_pQueryWidget, SIGNAL(lineRequested(const QString&, uint)),
		this, SLOT(slotQueryShowEditor(const QString&, uint)));

	// Run Cscope once the project has been loaded
	connect(m_pProjMgr, SIGNAL(projectOpened(const QString&)), this,
		SLOT(slotProjectOpened(const QString&)));

	// Refresh the file list if files were added to or removed from a project
	connect(m_pProjMgr, SIGNAL(fileListChanged()), this,
		SLOT(slotProjectFilesChanged()));

	// Add files to the file list when they are added to the project
	connect(m_pProjMgr, SIGNAL(filesAdded(const QStringList&)), this,
		SLOT(slotFilesAdded(const QStringList&)));

	// Store the new file tree root of a project
	connect(m_pFileView, SIGNAL(rootChanged(const QString&)), this,
		SLOT(slotProjectRootChanged(const QString&)));
	
	// Launch a "Find File" query when the "Find File..." button of the
	// file tree is clicked
	connect(m_pFileView, SIGNAL(findFile()), this, 
		SLOT(slotQueryFile()));

	// Store main window settings when closed
	setAutoSaveSettings();

	// If this is the first time the user has launched KScope, prompt him/her
	// to configure the global parameters
	if (Config().isFirstTime()) {
		KMessageBox::information(0, i18n("Welcome to KScope!\nIt appears to "
			"be the first time you are running this programme.\nPlease fill "
			"in the required configuration parameters."));
		slotConfigure();
		showMaximized(); // Use a maximised window the first time
	}
}

/**
 * Class destructor.
 */
KScope::~KScope()
{
	// Save configuration
	Config().store();
	Config().storeWorkspace(this);
	
	delete m_pEditMgr;
	delete m_pCscope;
	delete m_pProjMgr;
	delete m_pProgressDlg;
}

/**
 * Connects menu bar and toolbar commands with class members. These members
 * handle the actions associated with each command.
 */
void KScope::initActions()
{
	// File menu
	KStdAction::openNew(this, SLOT(slotNewFile()), actionCollection());
	KStdAction::open(this, SLOT(slotOpenFile()), actionCollection());
	KStdAction::close(this, SLOT(slotCloseEditor()), actionCollection());
	KStdAction::quit(this, SLOT(slotClose()), actionCollection());

	// Edit menu
	m_pExtEditAction = new KAction(i18n("Edit in E&xternal Editor"),
		KShortcut(QKeySequence("Ctrl+E")), this,
		SLOT(slotExtEdit()), actionCollection(), "edit_external_editor");

	// Project menu
	(void)new KAction(i18n("&New Project..."), 0, this, 
		SLOT(slotCreateProject()),	actionCollection(), "project_new");
	(void)new KAction(i18n("&Open Project..."), "project_open", 0, this,
		SLOT(slotOpenProject()), actionCollection(), "project_open");
	(void)new KAction(i18n("Add/Remove &Files..."), 0, this,
		SLOT(slotProjectFiles()), actionCollection(),
		"project_add_rem_files");
	(void)new KAction(i18n("&Properties..."), 0, this,
		SLOT(slotProjectProps()), actionCollection(),
		"project_properties");
	(void)new KAction(i18n("Open &Cscope.out..."), 0, this,
		SLOT(slotProjectCscopeOut()), actionCollection(),
		"project_cscope_out");
	(void)new KAction(i18n("&Close Project"), 0, this,
		SLOT(slotCloseProject()), actionCollection(), "project_close");

	// Cscope menu
	(void)new KAction(i18n("&References..."),
		KShortcut(QKeySequence("Ctrl+0")), this,
		SLOT(slotQueryReference()),	actionCollection(), "cscope_references");
	(void)new KAction(i18n("&Definition..."), 0,
		KShortcut(QKeySequence("Ctrl+1")), this,
		SLOT(slotQueryDefinition()), actionCollection(),
		"cscope_definition");
	(void)new KAction(i18n("&Called Functions..."), 0,
		KShortcut(QKeySequence("Ctrl+2")), this,
		SLOT(slotQueryCalled()), actionCollection(), "cscope_called");
	(void)new KAction(i18n("C&alling Functions..."), 0,
		KShortcut(QKeySequence("Ctrl+3")), this,
		SLOT(slotQueryCalling()), actionCollection(), "cscope_calling");
	(void)new KAction(i18n("Find &Text..."), 0,
		KShortcut(QKeySequence("Ctrl+4")), this,
		SLOT(slotQueryText()), actionCollection(), "cscope_text");
	(void)new KAction(i18n("Find &EGrep Pattern..."), 0,
		KShortcut(QKeySequence("Ctrl+6")), this,
		SLOT(slotQueryPattern()), actionCollection(), "cscope_pattern");
	(void)new KAction(i18n("Find &File..."), 0,
		KShortcut(QKeySequence("Ctrl+7")), this,
		SLOT(slotQueryFile()), actionCollection(), "cscope_file");
	(void)new KAction(i18n("Find &Including Files..."), 0,
		KShortcut(QKeySequence("Ctrl+8")), this,
		SLOT(slotQueryIncluding()), actionCollection(), "cscope_including");
	(void)new KAction(i18n("&Quick Definition"), 0,
		KShortcut(QKeySequence("Ctrl+]")), this,
		SLOT(slotQueryQuickDef()), actionCollection(), "cscope_quick_def");
	(void)new KAction(i18n("Ca&ll Tree..."), 0,
		KShortcut(QKeySequence("Ctrl+-")), this,
		SLOT(slotCallTree()), actionCollection(), "cscope_call_tree");
	(void)new KAction(i18n("Re&build database"),
		Pixmaps().getPixmap(KScopePixmaps::ToolbarReload), 0, this,
		SLOT(slotRebuildDB()), actionCollection(),
		"cscope_rebuild");
	(void)new KAction(i18n("Close &Query"), 0,
		0, this, SLOT(slotCloseQuery()), actionCollection(),
		"cscope_close_query");

	// Go menu
	(void)new KAction(i18n("&Previous Position"),
		Pixmaps().getPixmap(KScopePixmaps::GoBack), 
		KShortcut(QKeySequence(ALT + Key_Left)), this,
		SLOT(slotHistoryPrev()), actionCollection(),
		"go_prev_pos");
	(void)new KAction(i18n("&Next Position"),
		Pixmaps().getPixmap(KScopePixmaps::GoForward),
		KShortcut(QKeySequence(ALT + Key_Right)), this,
		SLOT(slotHistoryNext()), actionCollection(),
		"go_next_pos");
	(void)new KAction(i18n("Position &History..."), 0,
		0, this, SLOT(slotHistoryShow()), actionCollection(),
		"go_history");

	// View menu
	(void)new KAction(i18n("Show/Hide File List"),
		KShortcut(QKeySequence("Ctrl+/")), this,
		SLOT(slotToggleFileListDock()), actionCollection(),
		"view_toggle_filelist_dock");
	(void)new KAction(i18n("Show/Hide Query Window"),
		KShortcut(QKeySequence("Ctrl+.")), this,
		SLOT(slotToggleQueryDock()), actionCollection(),
		"view_toggle_query_dock");
	(void)new KAction(i18n("Show/Hide Query Buttons"),
		0, this, SLOT(slotToggleQueryButtons()), actionCollection(),
		"view_toggle_query_buttons");
	
	// Window menu
	(void)new KAction(i18n("Close &All Windows..."), 0, 0, this,
		SLOT(slotCloseAllWindows()), actionCollection(), "window_close_all");
		
	// Settings menu
	(void)new KAction(i18n("Configure &KScope..."), 0,
		0, this, SLOT(slotConfigure()), actionCollection(),
		"settings_kscope");

	// Create the context menu for the editor pages
	m_pEditorMenu->insertItem(i18n("&References..."), this,
		SLOT(slotQueryReference()));
	m_pEditorMenu->insertItem(i18n("&Definition..."), this,
		SLOT(slotQueryDefinition()));
	m_pEditorMenu->insertItem(i18n("&Called Functions..."), this,
		SLOT(slotQueryCalled()));
	m_pEditorMenu->insertItem(i18n("C&alling Functions..."), this,
		SLOT(slotQueryCalling()));
	m_pEditorMenu->insertItem(i18n("Find &Text..."), this,
		SLOT(slotQueryText()));
	m_pEditorMenu->insertItem(i18n("Find &EGrep Pattern..."), this,
		SLOT(slotQueryPattern()));
	m_pEditorMenu->insertItem(i18n("Find &File..."), this,
		SLOT(slotQueryFile()));
	m_pEditorMenu->insertItem(i18n("Find &Including Files..."), this,
		SLOT(slotQueryIncluding()));
	m_pEditorMenu->insertSeparator();
	m_pEditorMenu->insertItem(i18n("Quick Definition"), this,
		SLOT(slotQueryQuickDef()));
	m_pEditorMenu->insertItem(i18n("Ca&ll Tree..."), this,
		SLOT(slotCallTree()));
	m_pEditorMenu->insertSeparator();
	m_nExtEditId = m_pEditorMenu->insertItem(i18n("Edit in External Editor"),
		this, SLOT(slotExtEdit()));
	m_pEditorMenu->insertItem(i18n("Cl&ose Editor"), this,
		SLOT(slotCloseEditor()));
	
	m_pExtEditAction->setEnabled(Config().useExtEditor());
	m_pEditorMenu->setItemEnabled(m_nExtEditId, Config().useExtEditor());
}

/**
 * Creates the child widgets of this main window.
 */
void KScope::initMainWindow()
{
	KStatusBar* pStatus;
	KDockWidget* pMainDock;

	// Create the status bar
	pStatus = statusBar();
	pStatus->insertItem(i18n(" Line: N/A Col: N/A "), 0, 0, true);
	pStatus->insertItem(i18n(" Cscope: Not Running "), 1, 0, true);

	// Create the editors tab widget
	pMainDock = createDockWidget("Editors Window", QPixmap());
	m_pEditTabs = new EditorTabs(this, NULL,
		(QPopupMenu*)factory()->container("window", this));
	pMainDock->setWidget(m_pEditTabs);
	pMainDock->setDockSite(KDockWidget::DockCorner);

	setMainDockWidget(pMainDock);
	setView(pMainDock);
	pMainDock->setEnableDocking(KDockWidget::DockNone);
	
	// Create the query window
	m_pQueryDock = createDockWidget("Query Window", QPixmap());
	m_pQueryWidget = new QueryWidget(this);
	m_pQueryDock->setWidget(m_pQueryWidget);
	m_pQueryDock->manualDock(pMainDock, KDockWidget::DockBottom, 65);

	// Show the query window when a query completes
	connect(m_pQueryWidget, SIGNAL(queryFinished()), this, 
		SLOT(slotShowQueryWindow()));
	
	// Rerun queries when requested
	connect(m_pQueryWidget, SIGNAL(runQuery(uint, const QString&)), this,
		SLOT(slotQuery(uint, const QString&)));

	// Create the project window (file list)
	m_pFileViewDock = createDockWidget("File List Window", QPixmap());
	m_pFileView = new FileView(this);
	m_pFileViewDock->setWidget(m_pFileView);
	m_pFileViewDock->manualDock(pMainDock, KDockWidget::DockRight, 80);
	m_pFileList = m_pFileView->getFileList();
			
	// Create Cscope's error dialog
	m_pMsgDlg = new CscopeMsgDlg(this);
	
	// Restore dock configuration
	Config().loadWorkspace(this);
}

/**
 * Handles the "File->Quit" command. Closes the main window, which terminates
 * the application.
 */
void KScope::slotClose()
{
	// Destroy the main window
	KParts::DockMainWindow::close();
}

/**
 * Called when a request has been issued to close the main window.
 * Tries to close the active project.
 * @return	true if the main window can be closed, false otherwise
 */
bool KScope::queryClose()
{
	return slotCloseProject();
}

/**
 * Handles the "Project->New..." command.
 * Prompts the user for the name and folder for the project, and then creates
 * the project.
 */
void KScope::slotCreateProject()
{
	NewProjectDlg dlg(true, this);
	QStringList slTypes;

	// Prompt the user to close any active projects
	if (m_pProjMgr->isOpen()) {
		KMessageBox::error(0, i18n("Please close the active project before"
			" creating a new one"));
		return;
	}
	
	// Display the "New Project" dialog
	if (dlg.exec() != QDialog::Accepted)
		return;

	// Create the new project
	m_pProjMgr->create(dlg.getName(), dlg.getPath(), dlg.isKernel(),
		dlg.useInvIndex());
	m_pProjMgr->setFileTypes(dlg.getTypes());

	// Allow the user to add source files to the new project
	slotProjectFiles();
}

/**
 * Handles the "Project->Open..." command.
 * Prompts the user for a project file ("cscope.proj"), and opens the
 * selected project.
 */
void KScope::slotOpenProject()
{
	OpenProjectDlg dlg;

	if (dlg.exec() == QDialog::Rejected)
		return;

	openProject(dlg.m_pPathEdit->text());
}

/**
 * Handles the "Project->Add/Remove Files..." command.
 * Opens the project's files dialog, which allows the user to add and remove
 * source files.
 */
void KScope::slotProjectFiles()
{
	// A project must be open
	if (!m_pProjMgr->isOpen())
		return;

	// Display the files dialog
	ProjectFilesDlg dlg(m_pProjMgr, this);
	if (dlg.exec() == QDialog::Accepted)
		m_pProjMgr->writeList(&dlg);
}

/**
 * Handles the "Project->Properties..." command.
 * Opens the project's properties dialog, which allows the user to change
 * some attributes of the current project.
 * source files.
 */
void KScope::slotProjectProps()
{
	QStringList slTypes, slOptions;
	bool bKernel, bInvIndex;
	
	// A project must be open
	if (!m_pProjMgr->isOpen())
		return;

	// No properties for a temporary project
	if (m_pProjMgr->isTemporary()) {
		KMessageBox::error(0, i18n("The Project Properties dialogue is not "
			"available for temporary projects."));
		return;
	}
	
	// Create the properties dialog
	NewProjectDlg dlg(false, this);
	m_pProjMgr->getOptions(bKernel, bInvIndex);
	slTypes = QStringList::split(" ", m_pProjMgr->getFileTypes());
	dlg.setProperties(m_pProjMgr->getName(), m_pProjMgr->getPath(), 
		slTypes, bKernel, bInvIndex);
	
	// Display the properties dialog
	if (dlg.exec() != QDialog::Accepted)
		return;

	// Set new properties	
	m_pProjMgr->setFileTypes(dlg.getTypes());
	m_pProjMgr->setOptions(dlg.isKernel(), dlg.useInvIndex());
	
	// Cscope needs to be restarted if the Inverted Index option has changed
	if (bInvIndex != dlg.useInvIndex()) {
		m_pCscope->exit();
		m_pProjMgr->getOptions(slOptions);
		m_pCscope->run(m_pProjMgr->getPath(), slOptions);
	}
}

/**
 * Handles the "Cscope->Open Cscope.out..." menu command.
 * Prompts the user for a Cscope.out file, and, if successful, opens a new
 * session for working with this file.
 */
void KScope::slotProjectCscopeOut()
{
	QString sFilePath;
	
	// Prompt for a Cscope.out file
	sFilePath = KFileDialog::getOpenFileName();
	if (sFilePath.isEmpty())
		return;

	// Open a temporary project
	openCscopeOut(sFilePath);	
}

/**
 * Handles the "Cscope->References..." menu command.
 * Prompts the user for a symbol name, and initiates a query to find all 
 * references to that symbol.
 */
void KScope::slotQueryReference()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("References to a Symbol", sSymbol))
		slotQuery(CscopeFrontend::Reference, sSymbol);
}

/**
 * Handles the "Cscope->Definition..." menu command.
 * Prompts the user for a symbol name, and initiates a query to find the 
 * global definition of that symbol.
 */
void KScope::slotQueryDefinition()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Symbol Definition", sSymbol))
		slotQuery(CscopeFrontend::Definition, sSymbol);
}

/**
 * Handles the "Cscope->Called Functions..." menu command.
 * Prompts the user for a function name, and initiates a query to find all 
 * function calls from that function.
 */
void KScope::slotQueryCalled()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Functions Called by this Function", sSymbol))
		slotQuery(CscopeFrontend::Called, sSymbol);
}

/**
 * Handles the "Cscope->Calling Functions..." menu command.
 * Prompts the user for a function name, and initiates a query to find all 
 * functions calling that function.
 */
void KScope::slotQueryCalling()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Functions Calling this Function", sSymbol))
		slotQuery(CscopeFrontend::Calling, sSymbol);
}

/**
 * Handles the "Cscope->Find Text..." menu command.
 * Prompts the user for a string, and initiates a query to find all occurances 
 * of that string.
 */
void KScope::slotQueryText()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Search for text", sSymbol))
		slotQuery(CscopeFrontend::Text, sSymbol);
}

/**
 * Handles the "Cscope->Find EGrep Pattern..." menu command.
 * Prompts the user for a regular expression, and initiates a query to find 
 * all strings matching that pattern.
 */
void KScope::slotQueryPattern()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Search for EGrep pattern", sSymbol))
		slotQuery(CscopeFrontend::Pattern, sSymbol);
}

/**
 * Handles the "Cscope->Find File..." menu command.
 * Prompts the user for a file name, and initiates a query to find all files
 * having that name.
 */
void KScope::slotQueryFile()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Find Files Named", sSymbol))
		slotQuery(CscopeFrontend::FileName, sSymbol);
}

/**
 * Handles the "Cscope->Find Including Files..." menu command.
 * Prompts the user for a file name, and initiates a query to find all files
 * having an #include directive to that file.
 */
void KScope::slotQueryIncluding()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Files #including this file", sSymbol))
		slotQuery(CscopeFrontend::Including, sSymbol);
}

/**
 * Handles the "Cscope->Quick Definition" menu command.
 * Initiates a query to find the global definition of the symbol currently
 * selected or under the cursor. The user is prompted only if no symbol can
 * be found.
 */
void KScope::slotQueryQuickDef()
{
	QString sSymbol;

	// Prompt the user
	if (getSymbol("Symbol Definition", sSymbol, false))
		slotQuery(CscopeFrontend::Definition, sSymbol);
}

/**
 * Handles the "Cscope->Call Tree..." menu command.
 * Displays a tree of functions calling the requested function.
 */
void KScope::slotCallTree()
{
	QString sSymbol;

	if (getSymbol("Display a call tree", sSymbol)) {
		// Create a modeless call tree dialog
		CallTreeDlg* pCallTreeDlg = new CallTreeDlg(m_pCscope, sSymbol,
			this);
		pCallTreeDlg->show();

		// Open an editor whenever a function name is double-clicked
		connect(pCallTreeDlg, SIGNAL(lineRequested(const QString&, uint)),
			this, SLOT(slotShowEditor(const QString&, uint)));
	}
}

/**
 * Handles the "Cscope->Rebuild Database..." command.
 * Rebuilds Cscope's database for the current project.
 */
void KScope::slotRebuildDB()
{
	m_pCscope->rebuild();
}

/**
 * Handles the "Cscope->Close Query" command.
 * Closes the active query list.
 */
void KScope::slotCloseQuery()
{
	m_pQueryWidget->slotCloseCurrent();
}

/**
 * Handles the "Settings->Configure KScope..." command.
 * Displays the prferences dialog, which allows the user to set different
 * configuration parameters for KScope.
 */
void KScope::slotConfigure()
{
	PreferencesDlg dlg;

	// Apply the preferences if either the "Apply" or the "OK" buttons are
	// clicked
	connect(&dlg, SIGNAL(applyPref()), this, SLOT(slotApplyPref()));

	// Show the dialog
	dlg.exec();
}

/**
 * Completes the actions required when opening a project.
 * This slot is connected to the projectOpened() signal of the ProjectManager
 * object. The signal is emitted once a project has been successfully opened.
 * The function adds the project's files to the files list, and runs Cscope
 * on the project.
 * @param	sName	The opened project's name
 */
void KScope::slotProjectOpened(const QString& sName)
{
	QStringList slArgs;
	
	// Reflect the project's name in the title of the main window
	setCaption(sName);
	
	// Run Cscope only if the project has any files
	if (!m_pProjMgr->isEmpty()) {
		m_pProjMgr->fillList(m_pFileList);
		m_pProjMgr->getOptions(slArgs);
		m_pCscope->run(m_pProjMgr->getPath(), slArgs);
	}
}

/**
 * Refreshes the file list when files are added to or removed from a project,
 * and rebuilds the Cscope database.
 * This slot is attached to the fileListChanged() signal emitted by 
 * the ProjectManager object.
 */
void KScope::slotProjectFilesChanged()
{
	QStringList slArgs;
	
	// Refresh the file list
	m_pFileList->clear();
	m_pProjMgr->fillList(m_pFileList);
	
	// Rebuild the symbol database
	if (m_pCscope->isRunning()) {
		m_pCscope->rebuild();
	}
	else {
		// This may be the first time Cscope is run on this project, so we
		// need to pass the project's path and the command line arguments
		m_pProjMgr->getOptions(slArgs);
		m_pCscope->run(m_pProjMgr->getPath(), slArgs);
	}
}

/**
 * Adds a list of files to the current project.
 * This slot is connected to the filesAdded() signal of the ProjectManager
 * object. Once files are added to the project, they are also added to the
 * file list, and the project's database is rebuilt.
 * @param	slFiles	The list of file paths added to the project
 */
void KScope::slotFilesAdded(const QStringList& slFiles)
{
	QStringList::const_iterator itr;

	// Add the file paths to the project's file list
	for (itr = slFiles.begin(); itr != slFiles.end(); ++itr)
		m_pFileList->addItem(*itr);
	
	// Rebuild the database
	m_pCscope->rebuild();
}

/**
 * Starts a new Cscope query.
 * @param	nType	The numeric query type code
 * @param	sSymbol	The query's text, as entered by the user
 */
void KScope::slotQuery(uint nType, const QString& sSymbol)
{
	// Initiate the query	
	m_pQueryWidget->initQuery(nType, sSymbol);
	m_pCscope->query(nType, sSymbol, m_pQueryWidget->getActivePage());
}

/**
 * Opens a project.
 * If another project is currently active, it is cosed first.
 * @param	sDir	The directory of the project to open.
 */
void KScope::openProject(const QString& sDir)
{
	QStringList slOpenFiles;
	QStringList slQueryFiles;
	QStringList::Iterator itr;
	
	// Close the current project (may return false if the user clicks on the
	// "Cancel" button while prompted to save a file)
	if (!slotCloseProject())
		return;
	
	// Clear position history from previous project
	PosHistory().clear();

	// Open the project in the project manager
	m_pProjMgr->open(sDir);
	
	// Display the project's file dialog if no files are currently included
	// in the project
	if (m_pProjMgr->isEmpty())
		slotProjectFiles();
	
	// Show the file tree
	m_pFileView->setTreeRoot(m_pProjMgr->getRoot());
	
	// Restore the previous session
	m_pProjMgr->getOpenFiles(slOpenFiles);
	for (itr = slOpenFiles.begin(); itr != slOpenFiles.end(); ++itr) {
		QFile file(*itr);
		if (file.exists())
			slotAddEditor(*itr);
	}
	
	// Load previously stored queries
	m_pProjMgr->getQueryFiles(slQueryFiles);
	m_pQueryWidget->loadLockedQueries(m_pProjMgr->getPath(), slQueryFiles);
}

/**
 * Opens a temporary project for a Cscope.out file.
 * @param	The full path of the Cscope.out file
 */
void KScope::openCscopeOut(const QString& sFilePath)
{	
	// Close the current project (may return false if the user clicks on the
	// "Cancel" button while prompted to save a file)
	if (!slotCloseProject())
		return;

	// Open a temporary project for this cscope.out file
	m_pProjMgr->openCscopeOut(sFilePath);
}

/**
 * Shows/Hides the filelist dock.
 */
void KScope::slotToggleFileListDock()
{
	m_pFileViewDock->changeHideShowState();
}

/**
 * Shows/Hides the query dock.
 */
void KScope::slotToggleQueryDock()
{
	m_pQueryDock->changeHideShowState();
}

/**
 * Shows/Hides the query buttons.
 */
void KScope::slotToggleQueryButtons()
{
	m_pQueryWidget->showHideButtons();
}

/**
 * Closes the active project.
 * Closing a project involves closing all of the editor windows (prompting
 * the user for unsaved changes); terminating the Cscope process; and further
 * clean-up of the project's data.
 */
bool KScope::slotCloseProject()
{
	QStringList slOpenFiles;
	QStringList slQueryFiles;
	
	// Do nothing if no project is open
	if (!m_pProjMgr->isOpen())
		return true;
	
	// Close all open editor pages
	m_pEditTabs->getOpenFiles(slOpenFiles);
	if (!m_pEditTabs->removeAllPages())
		return false;

	// Store locked queries, and close all query pages
	m_pQueryWidget->saveLockedQueries(m_pProjMgr->getPath(), slQueryFiles);
	m_pQueryWidget->slotCloseAll();

	// Save current session
	m_pProjMgr->setOpenFiles(slOpenFiles);
	m_pProjMgr->setQueryFiles(slQueryFiles);
	
	// Close the project in the project manager, and terminate the Cscope
	// process
	m_pProjMgr->close();
	m_pCscope->exit();
	setCaption(QString::null);

	// Clear the contents of the file list
	m_pFileList->clear();

	return true;
}

/**
 * Handles the "Edit->Edit in External Editor" menu command.
 * Invokes an external editor for the current file and line number.
 */
void KScope::slotExtEdit()
{
	QString sCmdLine;
	KProcess proc;

	// Create the command line for the external editor	
	sCmdLine = Config().getExtEditor();
	sCmdLine.replace("%F", m_sCurFilePath);
	sCmdLine.replace("%L", QString::number(m_nCurLine));
	
	// Run the external editor
	proc.setUseShell(true);
	proc << sCmdLine;
	proc.start(KProcess::DontCare);
}

/**
 * Prompts the user for a symbol to query.
 * Shows a dialog with a line edit widget, where the user can enter a symbol
 * on which to query Cscope. The meaning of the symbol depends on the type of
 * query.
 * @param	sText	Determines the dialog's prompt text
 * @param	sSymbol	Holds the requested symbol, upon successful return
 * @param	bPrompt	If false, the user is prompted only if a symbol cannot be
 *					determined automatically
 * @return	true if the user hs enetered a symbol, false otherwise
 */
bool KScope::getSymbol(const QString& sText, QString& sSymbol, bool bPrompt)
{
	EditorPage* pPage;
	QString sSuggested;
	bool bAccepted;

	// Set the currently selected text, if any
	if ((pPage = m_pEditTabs->getCurrentPage()) != NULL)
		sSuggested = pPage->getSuggestedText();

	// Return if a symbol was found, and prompting is turned off
	if (!sSuggested.isEmpty() && !bPrompt) {
		sSymbol = sSuggested;
		return true;
	}
	
	// Show the dialog
	sSymbol = KInputDialog::getText(i18n("KScope Query"), sText, 
		sSuggested, &bAccepted, this);
	if (!bAccepted)
		return false;

	// Cannot accept empty strings
	sSymbol.stripWhiteSpace();
	return (!sSymbol.isEmpty());
}

/**
 * Opens a file in a new editor tab.
 * Creates a new editor page in a separate tab, and opens the requested file
 * in this page.
 * @param	sFileName	The file path to open
 */
void KScope::slotAddEditor(const QString& sFileName)
{
	KTextEditor::Document* pDoc;
	EditorPage* pPage;
	QString sFilePath;
	
	// If the file name is given using a relative path, we need to convert
	// it to an absolute one
	if (sFileName[0] != '/')
		sFilePath = m_pProjMgr->getPath() + "/" + sFileName;
	else
		sFilePath = sFileName;
	
	// Do not open a new editor if one exists for this file
	pPage = m_pEditTabs->findEditorPage(sFilePath);
	if (pPage != NULL)
		return;

	// Load a new document part
	pDoc = m_pEditMgr->add();
	if (pDoc == NULL)
		return;

	// Create the new editor page
	pPage = new EditorPage(pDoc, m_pEditorMenu, m_pEditTabs);
	m_pEditTabs->addEditorPage(pPage);

	connect(pPage, SIGNAL(fileOpened(EditorPage*, const QString&)), this,
		SLOT(slotFileOpened(EditorPage*, const QString&)));

	connect(pPage, SIGNAL(cursorPosChanged(uint, uint)), this,
		SLOT(slotShowCursorPos(uint, uint)));
	
	// Open the requested file
	pPage->open(sFilePath);
}

/*
 * Deletes an editor page after it has been removed.
 * The document object associated with the page is removed from the part 
 * manager, and the view object is removed from the GUI manager.
 * This slot is connected to the editorRemoved() signal of the EditorTabs 
 * object.
 * @param	pPage	The editor page to delete
 */
void KScope::slotDeleteEditor(EditorPage* pPage)
{
	guiFactory()->removeClient(pPage->getView());
	m_pEditMgr->remove(pPage->getDocument());
	delete pPage;
}

/**
 * Sets an editor part as active when its owner tab is chosen.
 * Whenever a different editor tab is chosen, its editor part should become
 * the active part. This means that this part's GUI is merged with the
 * application's, and that it responds to actions.
 * @param	pOldPage	The editor page that has ceased to be active
 * @param	pNewPage	The newly chosen editor page
 */
void KScope::slotChangeEditor(EditorPage* pOldPage, EditorPage* pNewPage)
{
	KXMLGUIFactory* pFactory = guiFactory();

	// Remove the current GUI
	if (pOldPage)
		pFactory->removeClient(pOldPage->getView());

	// Set the new active part and create its GUI
	if (pNewPage) {
		m_pEditMgr->setActivePart(pNewPage->getDocument());
		pFactory->addClient(pNewPage->getView());
		m_sCurFilePath = pNewPage->getFilePath();
		setCaption(m_pProjMgr->getName() + " - " + m_sCurFilePath);
	}
}

/**
 * Opens an editor for the given file and sets the cursor to the beginning of 
 * the requested line.
 * @param	sFilePath	The full path of the file to open for editing
 * @param	nLine		The number of the line on which to position the cursor
 */
void KScope::slotShowEditor(const QString& sFilePath, uint nLine)
{
	EditorPage* pPage;

	// Save current position in the position history
	PosHistory().add(m_sCurFilePath, m_nCurLine);
	
	// Open the requested file (or select an already-open editor page)
	slotAddEditor(sFilePath);
	pPage = m_pEditTabs->findEditorPage(sFilePath);
	if (pPage == NULL)
		return;

	// Set the cursor to the requested line
	pPage->slotGotoLine(nLine);
	
	// Add the new position to the position history
	PosHistory().add(sFilePath, nLine);
}

/**
 * A wrapper around slotShowEditor, that enables auto-hiding of the query
 * widget after a query result has been chosen.
 * This slot is connected to the lineRequested() signal emitted by a QueryPage
 * object.
 * @param	sFilePath	The full path of the file to open for editing
 * @param	nLine		The number of the line on which to position the cursor
 */
void KScope::slotQueryShowEditor(const QString& sFilePath, uint nLine)
{
	// Hide the query window, if it was hidden before a query was initiated
	if (m_bHideQueryOnSelection) {
		m_pQueryDock->changeHideShowState();
		m_bHideQueryOnSelection = false;
	}
	
	// Open an editor at the requested line
	slotShowEditor(sFilePath, nLine);
}

/*
 * Handles the "File->New" menu command.
 * Creates an editor page for a new unnamed file.
 */
void KScope::slotNewFile()
{
	KTextEditor::Document* pDoc;
	EditorPage* pPage;

	// Load a new document part
	pDoc = m_pEditMgr->add();
	if (pDoc == NULL)
		return;

	// Create the new editor page
	pPage = new EditorPage(pDoc, m_pEditorMenu, m_pEditTabs);
	m_pEditTabs->addEditorPage(pPage);
	
	// Mark the page as containing a new file
	pPage->setNewFile();
}

/*
 * Handles the "File->Open" menu command.
 * Prompts the user for a file name, and opens it in a new editor page.
 */
void KScope::slotOpenFile()
{
	QString sFileName;
	
	sFileName = KFileDialog::getOpenFileName();
	if (!sFileName.isEmpty())
		slotAddEditor(sFileName);
}

/**
 * Handles the "File->Close" menu command.
 * Closes the currently active editor page.
 */
void KScope::slotCloseEditor()
{
	m_pEditTabs->removeCurrentPage();
}

/*
 * Handles the "Window->Close All" menu command.
 * Closes all open editor pages.
 */
void KScope::slotCloseAllWindows()
{
	m_pEditTabs->removeAllPages();
}

/**
 * Handles the "Go->Back" menu command.
 * Moves to the previous position in the position history.
 */
void KScope::slotHistoryPrev()
{
	QString sFilePath;
	uint nLine;
	
	if (PosHistory().prev(sFilePath, nLine)) {
		PosHistory().accept(false);
		slotShowEditor(sFilePath, nLine);
		PosHistory().accept(true);
	}
}

/**
 * Handles the "Go->Forward" menu command.
 * Moves to the previous position in the position history.
 */
void KScope::slotHistoryNext()
{
	QString sFilePath;
	uint nLine;
	
	if (PosHistory().next(sFilePath, nLine)) {
		PosHistory().accept(false);
		slotShowEditor(sFilePath, nLine);
		PosHistory().accept(true);
	}
}

/**
 * Handles the "Go->Position History..." menu command.
 * Displays a dialog with all recorded positions the user has moved to.
 */
void KScope::slotHistoryShow()
{
	PositionHistoryDlg dlg(this);
	QString sFilePath;
	uint nLine;
	
	if (dlg.exec() == QDialog::Accepted && 
		PosHistory().current(sFilePath, nLine)) {
		slotShowEditor(sFilePath, nLine);
	}
}

/**
 * Reports whether Cscope is running.
 * @param	bRunning	true if Cscope is running, false otherwise
 */
void KScope::slotCscopeRunning(bool bRunning)
{
	KStatusBar* pStatus = statusBar();

	if (!bRunning) {
		pStatus->changeItem(i18n(" Cscope: Not Running "), 1);
		delete m_pProgressDlg;
		m_pProgressDlg = NULL;
	}
}

/**
 * Reports whether Cscope is working or ready for new queries.
 * @param	bReady	true is Cscope is ready, false otherwise
 */
void KScope::slotCscopeState(bool bReady)
{
	KStatusBar* pStatus = statusBar();

	if (bReady) {
		pStatus->changeItem(i18n(" Cscope: Ready "), 1);
		delete m_pProgressDlg;
		m_pProgressDlg = NULL;
	}
	else {
		pStatus->changeItem(i18n(" Cscope: Working "), 1);
	}
}

/**
 * Displays a progress dialog for a lengthy Cscope action.
 * This slot is connected to the progress() signal emitted by the
 * CscopeFrontend object.
 * @param	sLabel	The text to display in the dialog
 * @param	nValue	The new value of the progress bar
 */
void KScope::slotProgress(const QString& sLabel, int nValue)
{
	// If the dialog does not exist, create it
	if (m_pProgressDlg == NULL) {
		m_pProgressDlg = new ProgressDlg;
		m_pProgressDlg->setMinimumDuration(500);

		// Clicking "Cancel" should stop the Cscope operation
		connect(m_pProgressDlg, SIGNAL(cancelClicked()), this,
			SLOT(slotCancelCscope()));
	}

	// Create a new progress dialog
	m_pProgressDlg->setLabel(sLabel);
	m_pProgressDlg->setValue(nValue);

	// If the value has reached 100, disable the "Cancel" button
	if (nValue >= 100)
		m_pProgressDlg->setAllowCancel(false);
}

/**
 * Sends an interrupt signal to a running Cscope process.
 * This slot is connceted to the "Cancel" button of the progress dialog.
 */
void KScope::slotCancelCscope()
{
	m_pCscope->slotCancel();
	m_pProgressDlg = NULL;
}

/**
 * Applies the selected user preferences once the "Apply" or "OK" buttons in
 * the preferences dialog is clicked.
 */
void KScope::slotApplyPref()
{
	m_pQueryWidget->applyPrefs();
	m_pFileList->applyPrefs();
	m_pEditTabs->applyPrefs();
	
	// Enable/disable the external editor menu item
	m_pExtEditAction->setEnabled(Config().useExtEditor());
	m_pEditorMenu->setItemEnabled(m_nExtEditId, Config().useExtEditor());
}

/**
 * Displays the query window once a query has completed (in case it is not
 * already visible.)
 * If the window is not visible, this information is saved and the query
 * window is hidden again when the user makes a record selection.
 */
void KScope::slotShowQueryWindow()
{
	// Show the query window, if currently hidden
	m_bHideQueryOnSelection = m_pQueryDock->isHidden();
	m_pQueryDock->makeDockVisible();
}

/**
 * Displays the current cursor position, whenever it is moved by the user.
 * This slot is connected to the cursorPosChanged() signal emitted by an
 * EditorPage object.
 * @param	nLine	The new line number
 * @param	nCol	The new column number
 */
void KScope::slotShowCursorPos(uint nLine, uint nCol)
{
	KStatusBar* pStatus = statusBar();
	QString sText;
	
	/* Show the line and column numbers. */
	QTextOStream(&sText) << " Line: " << nLine + 1 << " Col: " << nCol + 1
		<< " ";
	pStatus->changeItem(sText, 0);
	
	/* Store the current line. */
	m_nCurLine = nLine + 1;
}

/**
 * Stores the path of a newly opened file.
 * This slot is connected to the fileOpened() signal emitted by an
 * EditorPage object.
 * @param				The page that emitted the signal (ignored)
 * @param	sFilePath	The full path of the opened file
 */
void KScope::slotFileOpened(EditorPage*, const QString& sFilePath)
{
	m_sCurFilePath = sFilePath;
	setCaption(m_pProjMgr->getName() + " - " + m_sCurFilePath);
}

/**
 * Sets a new file tree root for the current project.
 * This slot is connected to the rootChanged() signal of the FileView object.
 * @param	sRoot	The full path of the directory to serve as the new root
 */
void KScope::slotProjectRootChanged(const QString& sRoot)
{
	m_pProjMgr->setRoot(sRoot);
}
